/*
 * Created on 05/03/2008
 */
package numerics;

import objectDefinitions.QuadCoef;
import objectDefinitions.Ray;

/**
 * @author Marlus Barbosa de Souza - mbs2
 * @author Artur Lira dos Santos - als3
 */

//From quadric.cpp
/*--------------------------------------------------------------------------*/
/*	QUADRIC.C																*/
/*		This module contains the routine (and helper routines) to compute	*/
/*		the point of intersecion between a ray and a quadric surface		*/
/*																			*/
/*- Modification History ---------------------------------------------------*/
/*	When:		Who:					Comments:							*/
/*																			*/
/*	??-???-??	Victor Klassen			Initial implementation				*/
/*	03-Sep-95	Christopher G. Healey	Combined code into a single module	*/
/*	02-Mar-2007	Artur Lira				Translation C -> Java				*/
/*	03-Mar-2007 Marlus Souza			Revised the code translation		*/
/*	07-Mar-2007 Marlus Souza			Added some functions for the		*/
/*										RayCife project 					*/
/*--------------------------------------------------------------------------*/

//Classe para resolver a interseco
public class Quadric {
	
	//Compute the intersection point, if it exists, between the given ray and
	//the given object
	//
	//Victor uses a strange formula from Watt & Watt for quadrics:
	//
	//		Ax^2 + Ey^2 + Hz^2 + Bxy + Fyz + Cxz + Dx + Gy + Jz + K
	//
	//rather than the standard formula:
	//
	//		Ax^2 + By^2 + Cz^2 + 2Dxy + 2Eyz + 2Fxz + 2Gx + 2Hy + 2Jz + K
	//
	//	Ray arg0: Ray being shot into scene
	//	Quad arg1: Object to test for intersection
	
	public static double intersect(Ray ray, QuadCoef coefficients) {
		
		double a, b, c, d, e;			//Coefficents of equation of..
		double f, g, h, j, k;			//..quadric surface
		double acoef, bcoef, ccoef;		//Intersection coefficents
		double dx, dy, dz; 				//Direction - origin coordinates
		double disc;					//Distance to intersection
		double root;					//Root of distance to intersection
		double t;						//Distance along ray to intersection
		double x0, y0, z0;				//Origin coordinates
		
		a = coefficients.getA();
		b = coefficients.getB();
		c = coefficients.getC();
		d = coefficients.getD();
		e = coefficients.getE();
		f = coefficients.getF();
		g = coefficients.getG();
		h = coefficients.getH();
		j = coefficients.getJ();
		k = coefficients.getK();
		
		x0 = ray.getOrig().getX();
		y0 = ray.getOrig().getY();
		z0 = ray.getOrig().getZ();
		
		dx = ray.getDir().getX() - x0;
		dy = ray.getDir().getY() - y0;
		dz = ray.getDir().getZ() - z0;
		
		acoef = 2 * f * dx * dz + 2 * e * dy * dz + c * dz * dz + b * dy * dy +
			a * dx * dx + 2 * d * dx * dy;
		
		bcoef = 2 * b * y0 * dy + 2 * a * x0 * dx + 2 * c * z0 * dz +
			2 * h * dy + 2 * g * dx + 2 * j * dz + 2 * d * x0 * dy +
			2 * e * y0 * dz + 2 * e * z0 * dy + 2 * d * y0 * dx +
			2 * f * x0 * dz + 2 * f * z0 * dx;
		
		ccoef = a * x0 * x0 + 2 * g * x0 + 2 * f * x0 * z0 + b * y0 * y0 +
			2 * e * y0 * z0 + 2 * d * x0 * y0 + c * z0 * z0 + 2 * h * y0 +
			2 * j * z0 + k;
		
		//The following was modified by David J. Brandow to allow for planar
		//quadrics
		if(1.0 + acoef == 1.0){
			if(1.0 + bcoef == 1.0){
				return -1.0;
			}
			t = (-ccoef) / (bcoef);
		}else{
			disc = bcoef * bcoef - 4 * acoef * ccoef;
			if(1.0 + disc < 1.0){
				return -1.0;
			}
			root = Math.sqrt(disc);
			t = (-bcoef - root)/(acoef + acoef);
			if(t < 0.0){
				t = (-bcoef + root) / (acoef + acoef);
			}
		}
		
		if(t < 0.001){
			return -1.0;
		}
		return t;
	}							//End procedure intersect
}